use state::Event;
use state::Surface;
use driver;
use std::sync::mpsc::Sender;
use gfx_device_gl;
use gfx_window_glutin;
use gfx::traits::FactoryExt;
use gfx::{self, Device};
use gfx::format::*;
use glutin::{
    self,
    event::{
        Event as GEvent,
        VirtualKeyCode,
        MouseButton,
        MouseScrollDelta,
        TouchPhase,
        WindowEvent::{
            self,
            //KeyboardInput,
            CloseRequested,
            Resized,
            MouseWheel,
            MouseInput,
            CursorMoved,
            ReceivedCharacter,
        },
        KeyboardInput,
        ElementState::Pressed,
    },
    event_loop::EventLoop,
    WindowedContext,
    PossiblyCurrent,
};
use cgmath::{Matrix2, Matrix3, Vector2, Vector3, SquareMatrix, Rad, Zero};
use imgui::{
    Context,
    Condition,
    im_str,
    FontGlyphRanges,
    Ui,
    FontConfig,
    Window,
};
use imgui_gfx_renderer::{Renderer, Shaders};
use image;
use proto::{ObjXY,ObjID,ResID};
use std::collections::{HashMap, BTreeMap};

mod delta {
    use std::time::Instant;

    pub struct Delta {
        last: Instant
    }

    impl Delta {
        pub fn new () -> Delta {
            Delta { last: Instant::now() }
        }

        pub fn tick (&mut self) -> f32 {
            let now = Instant::now();
            let delta = now - self.last;
            self.last = now;
            delta.as_secs() as f32 + delta.subsec_nanos() as f32 / 1_000_000_000.0
        }
    }
}

const BLACK: [f32; 4] = [0.0, 0.0, 0.0, 1.0];

gfx_defines! {
    vertex VertexCol {
        position: [f32; 2] = "position",
        color: [f32; 3] = "color",
        height: i32 = "height",
    }

    vertex VertexTex {
        position: [f32; 2] = "position",
        uv: [f32;2] = "uv",
    }

    pipeline pipe_col {
        vbuf: gfx::VertexBuffer<VertexCol> = (),
        transform: gfx::Global<[[f32; 3]; 3]> = "transform",
        threshold: gfx::Global<i32> = "threshold",
        //out: gfx::RenderTarget<Rgba8> = "Target0",
        out: gfx::BlendTarget<Rgba8> = ("Target0", gfx::state::ColorMask::all(), gfx::preset::blend::ALPHA),
    }

    pipeline pipe_tex {
        vbuf: gfx::VertexBuffer<VertexTex> = (),
        tex: gfx::TextureSampler<[f32; 4]> = "sampler",
        transform: gfx::Global<[[f32; 3]; 3]> = "transform",
        //out: gfx::RenderTarget<Rgba8> = "Target0",
        out: gfx::BlendTarget<Rgba8> = ("Target0", gfx::state::ColorMask::all(), gfx::preset::blend::ALPHA),
    }
}

//FIXME transparent color is BLACK for some reason
const VERTEX_SHADER_COL: &str = "
    #version 130

    in vec2 position;
    in vec3 color;
    in int height;
    out vec4 v_color;

    uniform mat3 transform;
    uniform int threshold;

    void main() {
        if (height > threshold) {
            v_color = vec4(color, 1.0);
        } else {
            v_color = vec4(0.0, 0.0, 0.0, 0.0);
        }
        gl_Position = vec4(transform * vec3(position, 1.0), 1.0);
    }
";

const FRAGMENT_SHADER_COL: &str = "
    #version 130

    in vec4 v_color;
    out vec4 Target0;

    void main() {
        Target0 = v_color;
    }
";

const VERTEX_SHADER_TEX: &str = "
    #version 130

    in vec2 position;
    in vec2 uv;
    out vec2 v_uv;

    uniform mat3 transform;

    void main() {
        v_uv = uv;
        gl_Position = vec4(transform * vec3(position, 1.0), 1.0);
    }
";

const FRAGMENT_SHADER_TEX: &str = "
    #version 130

    in vec2 v_uv;
    out vec4 Target0;

    uniform sampler2D sampler;

    void main() {
        Target0 = texture(sampler, v_uv);
    }
";

impl VertexCol {
    fn new (x: f32, y: f32, color: [f32; 3], height: i32) -> VertexCol {
        VertexCol {
            position: [x, y],
            color: color,
            height: height
        }
    }

    fn shift (&mut self, x: f32, y: f32) {
        self.position[0] += x;
        self.position[1] += y;
    }
}
impl VertexTex {
    fn shift (&mut self, x: f32, y: f32) {
        self.position[0] += x;
        self.position[1] += y;
    }
}

struct ObjCol {
    vertices: Vec<VertexCol>,
    indices: Vec<u32>,
}

impl ObjCol {

    fn new () -> ObjCol {
        ObjCol {
            vertices: Vec::new(),
            indices: Vec::new()
        }
    }

    fn with_capacity (vcap: usize, icap: usize) -> ObjCol {
        ObjCol {
            vertices: Vec::with_capacity(vcap),
            indices: Vec::with_capacity(icap)
        }
    }

    fn translate (&mut self, x: f32, y: f32) {
        for v in self.vertices.iter_mut() {
            v.shift(x,y);
        }
    }

    #[allow(dead_code)]
    fn grid_from_heights (size: usize, cell_size: f32, grid_x: i32, grid_y: i32, heights: &[i16], t/*thickness*/: f32) -> ObjCol {
        assert_eq!(heights.len(), size*size);
        const WARM: [f32; 3] = [1.0, 0.3, 0.3];
        const COLD: [f32; 3] = [0.3, 0.3, 1.0];
        let mut obj = ObjCol::new();
        let mut i = 0;
        for x in 0..size-1 {
            for y in 0..size-1 {
                let a = heights[y*size+x];
                let b = heights[y*size+x+1];
                let c = heights[(y+1)*size+x];
                if a != b {
                    let ax = x as f32 * cell_size;// + 0.5;
                    let ay = y as f32 * cell_size;
                    let bx = ax + cell_size;// - 1.0;
                    if a > b {
                        let height = (a - b) as i32;
                        obj.vertices.push(VertexCol::new(ax, ay+t, WARM, height));
                        obj.vertices.push(VertexCol::new(bx, ay,   WARM, height));
                        obj.vertices.push(VertexCol::new(ax, ay-t, WARM, height));
                    } else {
                        let height = (b - a) as i32;
                        obj.vertices.push(VertexCol::new(ax, ay,   COLD, height));
                        obj.vertices.push(VertexCol::new(bx, ay+t, COLD, height));
                        obj.vertices.push(VertexCol::new(bx, ay-t, COLD, height));
                    }
                    obj.indices.push(i); i+=1;
                    obj.indices.push(i); i+=1;
                    obj.indices.push(i); i+=1;
                }
                if a != c {
                    let ax = x as f32 * cell_size;
                    let ay = y as f32 * cell_size;// + 0.5;
                    let cy = ay + cell_size;// - 1.0;
                    if a > c {
                        let height = (a - b) as i32;
                        obj.vertices.push(VertexCol::new(ax-t, ay, WARM, height));
                        obj.vertices.push(VertexCol::new(ax, cy,   WARM, height));
                        obj.vertices.push(VertexCol::new(ax+t, ay, WARM, height));
                    } else {
                        let height = (b - a) as i32;
                        obj.vertices.push(VertexCol::new(ax, ay,   COLD, height));
                        obj.vertices.push(VertexCol::new(ax-t, cy, COLD, height));
                        obj.vertices.push(VertexCol::new(ax+t, cy, COLD, height));
                    }
                    obj.indices.push(i); i+=1;
                    obj.indices.push(i); i+=1;
                    obj.indices.push(i); i+=1;
                }
            }
        }
        obj.translate(size as f32 * cell_size * grid_x as f32, size as f32 * cell_size * grid_y as f32);
        obj
    }

    #[allow(dead_code)]
    fn grid_from_heights2 (size: usize, cell_size: f32, grid_x: i32, grid_y: i32, heights: &[i16]) -> ObjCol {
        use std::cmp::max;
        assert_eq!(heights.len(), size*size);
        let mut obj = ObjCol::new();
        for x in 0..size-1 {
            for y in 0..size-1 {
                let a = heights[y*size+x];
                let b = heights[y*size+x+1];
                let c = heights[(y+1)*size+x];
                let d = heights[(y+1)*size+x+1];
                let delta_ab = if a > b { a - b } else { b - a } as usize;
                let delta_ac = if a > c { a - c } else { c - a } as usize;
                let delta_bd = if b > d { b - d } else { d - b } as usize;
                let delta_cd = if c > d { c - d } else { d - c } as usize;
                let delta = max(max(delta_ab, delta_ac), max(delta_cd, delta_bd));
                if delta > 0 {
                    let ax = x as f32 * cell_size + 2.0;
                    let ay = y as f32 * cell_size + 2.0;
                    let dx = ax + cell_size - 4.0;
                    let dy = ay + cell_size - 4.0;
                    obj.add_quad(ax, ay, dx, dy, [0.1,0.1,0.1], delta as i32);
                } 
            }
        }
        obj.translate(size as f32 * cell_size * grid_x as f32, size as f32 * cell_size * grid_y as f32);
        obj
    }

    fn add_quad (&mut self, x1: f32, y1: f32, x2: f32, y2: f32, color: [f32; 3], height: i32) {
        let len = self.vertices.len() as u32;
        self.vertices.push(VertexCol::new(x1, y1, color, height));
        self.vertices.push(VertexCol::new(x1, y2, color, height));
        self.vertices.push(VertexCol::new(x2, y2, color, height));
        self.vertices.push(VertexCol::new(x2, y1, color, height));
        self.indices.push(len);
        self.indices.push(len+1);
        self.indices.push(len+2);
        self.indices.push(len);
        self.indices.push(len+2);
        self.indices.push(len+3);
    }

    fn from_objects (objects: &BTreeMap<ObjID,(ObjXY,ResID)>, hero_x: f32, hero_y: f32, hf_x: f32, hf_y: f32) -> ObjCol {
        let mut obj = ObjCol::with_capacity(objects.len()*4, objects.len()*6);
        for &(ObjXY(x,y),_resid) in objects.values() {
            obj.add_quad(x as f32 - 2.0, y as f32 - 2.0, x as f32 + 2.0, y as f32 + 2.0, [1.0,1.0,1.0], 1);
        }
        obj.add_quad(hero_x - 2.0, hero_y - 2.0, hero_x + 2.0, hero_y + 2.0, [0.1,1.0,0.1], 1);
        obj.add_quad(hf_x - 2.0, hf_y - 2.0, hf_x + 2.0, hf_y + 2.0, [0.1,1.0,1.0], 1);
        obj
    }

    fn bake (self, main_color: gfx::handle::RenderTargetView<gfx_device_gl::Resources, Rgba8>, factory: &mut gfx_device_gl::Factory, threshold: usize) -> BakedObjCol {
        let (vertex_buffer, slice) = factory.create_vertex_buffer_with_slice(&self.vertices, &self.indices[..]);
        let data = pipe_col::Data {
            vbuf: vertex_buffer,
            transform: Matrix3::identity().into(),
            threshold: threshold as i32,
            out: main_color
        };
        BakedObjCol {
            slice: slice,
            data: data
        }
    }
}

struct ObjTex {
    vertices: Vec<VertexTex>,
    indices: Vec<u32>,
    image: image::ImageBuffer<image::Rgba<u8>,Vec<u8>>,
}

impl ObjTex {
    fn with_capacity (vcap: usize, icap: usize, image_x: usize, image_y: usize) -> ObjTex {
        ObjTex {
            vertices: Vec::with_capacity(vcap),
            indices: Vec::with_capacity(icap),
            image: image::ImageBuffer::new(image_x as u32, image_y as u32),
        }
    }

    fn translate (&mut self, x: f32, y: f32) {
        for v in self.vertices.iter_mut() {
            v.shift(x,y);
        }
    }

    fn plane <F> (size: f32, color_fn: F) -> ObjTex where F: Fn(usize,usize)->[u8; 4] {
        let mut obj = ObjTex::with_capacity(4, 6, 100, 100);
        for y in 0..100 {
            for x in 0..100 {
                obj.image.put_pixel(x, y, image::Rgba(color_fn(x as usize, y as usize)));
            }
        }
        obj.vertices.push(VertexTex{position:[0.0, 0.0], uv:[0.0, 0.0] });
        obj.vertices.push(VertexTex{position:[size, 0.0], uv:[1.0, 0.0] });
        obj.vertices.push(VertexTex{position:[size, size], uv:[1.0, 1.0] });
        obj.vertices.push(VertexTex{position:[0.0, size], uv:[0.0, 1.0] });
        obj.indices.push(0);
        obj.indices.push(1);
        obj.indices.push(2);
        obj.indices.push(0);
        obj.indices.push(2);
        obj.indices.push(3);
        obj
    }

    fn plane_from_owning (size: f32, grid_x: i32, grid_y: i32, owning: &[u8]) -> ObjTex {
        assert_eq!(owning.len(), 100*100);
        let mut obj = ObjTex::plane(size, |x,y|{
            let owning = owning[y*100+x];
            if owning == 0 {
                [0,0,0,0]
            } else {
                //TODO different colors for personal and city claims
                [127*(owning&1)      + 127*((owning>>1)&1),
                 127*((owning>>2)&1) + 127*((owning>>3)&1),
                 127*((owning>>4)&1) + 127*((owning>>5)&1), 63]
            }
        });
        obj.translate(size as f32 * grid_x as f32, size as f32 * grid_y as f32);
        obj
            /*
               if owning[y*100+x] == 0 {
               palette[index]
               } else {
               let r: u8 = palette[index][0];
               let g = palette[index][1];
               let b = palette[index][2];
               let a = palette[index][3];
               [r.saturating_add(50u8),g,b,a]
               }
               */
    }

    fn plane_from_tiles (size: f32, grid_x: i32, grid_y: i32, tiles: &[u8], palette: &[[u8; 4]; 256]) -> ObjTex {
        assert_eq!(tiles.len(), 100*100);
        let mut obj = ObjTex::plane(size, |x,y|{
            let index = tiles[y*100+x] as usize;
            palette[index]
        });
        obj.translate(size as f32 * grid_x as f32, size as f32 * grid_y as f32);
        obj
    }

    fn bake (self, main_color: gfx::handle::RenderTargetView<gfx_device_gl::Resources, Rgba8>, factory: &mut gfx_device_gl::Factory) -> BakedObjTex {
        use gfx::Factory;
        use gfx::texture::{SamplerInfo, FilterMethod, WrapMode, Lod, PackedColor};
        use gfx::state::Comparison;
        let (vertex_buffer, slice) = factory.create_vertex_buffer_with_slice(&self.vertices, &self.indices[..]);
        //TODO find more convenient way to create "nearest" sampler
        let sampler = factory.create_sampler(
            SamplerInfo{
                filter: FilterMethod::Scale,
                wrap_mode: (WrapMode::Tile, WrapMode::Tile, WrapMode::Tile),
                lod_bias: Lod::from(0.0),
                lod_range: (Lod::from(0.0), Lod::from(0.0)),
                comparison: Some(Comparison::Never),
                border: PackedColor(0),
            }
            );
        let (width, height) = (100,100); //FIXME self.image.dimensions();
        let kind = gfx::texture::Kind::D2(width as u16, height as u16, gfx::texture::AaMode::Single);
        let mipmap = gfx::texture::Mipmap::Allocated;
        let (_, view) = factory.create_texture_immutable_u8::<Rgba8>(kind, mipmap, &[&self.image]).unwrap();
        let data = pipe_tex::Data {
            vbuf: vertex_buffer,
            tex: (view, sampler),
            transform: Matrix3::identity().into(),
            out: main_color
        };
        BakedObjTex {
            slice: slice,
            data: data
        }
    }
}

struct BakedObjCol {
    slice: gfx::Slice<gfx_device_gl::Resources>,
    data: pipe_col::Data<gfx_device_gl::Resources>
}

struct BakedObjTex {
    slice: gfx::Slice<gfx_device_gl::Resources>,
    data: pipe_tex::Data<gfx_device_gl::Resources>
}

fn run_ui <'a> (ui: &Ui<'a>,
                fps: usize,
                threshold: usize,
                show_tiles: &mut bool,
                show_heights: &mut bool,
                show_owning: &mut bool,
                v1: &mut i32,
                objects: &BTreeMap<ObjID,(ObjXY,ResID)>,
                resources: &BTreeMap<ResID,String>) -> bool {
    Window::new(im_str!("Клёцка"))
        .size([300.0, 600.0], Condition::FirstUseEver)
        .position([10.0, 10.0], Condition::FirstUseEver)
        .build(ui, || {
            ui.text(im_str!("Привет, Мир!!!"));
            ui.separator();
            let [x,y] = ui.io().mouse_pos;
            ui.text(im_str!("Mouse Position: ({:.1},{:.1})", x, y));
            ui.text(im_str!("FPS: {}", fps));
            ui.text(im_str!("threshold: {}", threshold));
            ui.checkbox(im_str!("tiles"), show_tiles);
            ui.checkbox(im_str!("heights"), show_heights);
            /*TODO indent*/ ui.radio_button(im_str!("quads"), v1, 0);
            /*TODO indent*/ ui.radio_button(im_str!("arrows"), v1, 1);
            /*TODO indent*/ ui.radio_button(im_str!("heatmap"), v1, 2);
            ui.checkbox(im_str!("owning"), show_owning);
        });
    Window::new(im_str!("Объекты"))
    //ui.window(im_str!("Объекты: {}", objects.len()))
        .size([300.0, 600.0], Condition::FirstUseEver)
        .position([300.0, 10.0], Condition::FirstUseEver)
        .build(ui, || {
            for ( &objid, &(ObjXY(_x,_y), ref resid) ) in objects.iter() {
                let resname = match resources.get(resid) {
                    Some(ref name) => name,
                    None => "???"
                };
                ui.text(im_str!("{} {}", objid, resname));
            }
        });

    true
}

#[derive(Copy, Clone, PartialEq, Debug, Default)]
struct MouseState {
    pos: (i32, i32),
    pressed: (bool, bool, bool),
    wheel: f32,
}

fn ui_handle_event (imgui: &mut Context, mouse_state: &mut MouseState, event: &GEvent<()>) {
    if let GEvent::WindowEvent { ref event, .. } = *event {
        match *event {
            //Resized(_, _) => {
            //gfx_window_glutin::update_views(&window, &mut main_color, &mut main_depth);
            //renderer.update_render_target(main_color.clone());
            //}
            //Closed => quit = true,
            WindowEvent::KeyboardInput { input, .. } => {
                let pressed = input.state == Pressed;
                match input.virtual_keycode {
                    Some(VirtualKeyCode::Tab) => imgui.io_mut().key_map[0] = pressed as u32, //FIXME investigate how this mapping is working now
                    Some(VirtualKeyCode::Left) => imgui.io_mut().key_map[1] = pressed as u32, //FIXME investigate how this mapping is working now
                    Some(VirtualKeyCode::Right) => imgui.io_mut().key_map[2] = pressed as u32, //FIXME investigate how this mapping is working now
                    Some(VirtualKeyCode::Up) => imgui.io_mut().key_map[3] = pressed as u32, //FIXME investigate how this mapping is working now
                    Some(VirtualKeyCode::Down) => imgui.io_mut().key_map[4] = pressed as u32, //FIXME investigate how this mapping is working now
                    Some(VirtualKeyCode::PageUp) => imgui.io_mut().key_map[5] = pressed as u32, //FIXME investigate how this mapping is working now
                    Some(VirtualKeyCode::PageDown) => imgui.io_mut().key_map[6] = pressed as u32, //FIXME investigate how this mapping is working now
                    Some(VirtualKeyCode::Home) => imgui.io_mut().key_map[7] = pressed as u32, //FIXME investigate how this mapping is working now
                    Some(VirtualKeyCode::End) => imgui.io_mut().key_map[8] = pressed as u32, //FIXME investigate how this mapping is working now
                    Some(VirtualKeyCode::Delete) => imgui.io_mut().key_map[9] = pressed as u32, //FIXME investigate how this mapping is working now
                    Some(VirtualKeyCode::Back) => imgui.io_mut().key_map[1] = pressed as u32, //FIXME investigate how this mapping is working now
                    Some(VirtualKeyCode::Return) => imgui.io_mut().key_map[1] = pressed as u32, //FIXME investigate how this mapping is working now
                    Some(VirtualKeyCode::Escape) => imgui.io_mut().key_map[1] = pressed as u32, //FIXME investigate how this mapping is working now
                    Some(VirtualKeyCode::A) => imgui.io_mut().key_map[1] = pressed as u32, //FIXME investigate how this mapping is working now
                    Some(VirtualKeyCode::C) => imgui.io_mut().key_map[1] = pressed as u32, //FIXME investigate how this mapping is working now
                    Some(VirtualKeyCode::V) => imgui.io_mut().key_map[1] = pressed as u32, //FIXME investigate how this mapping is working now
                    Some(VirtualKeyCode::X) => imgui.io_mut().key_map[1] = pressed as u32, //FIXME investigate how this mapping is working now
                    Some(VirtualKeyCode::Y) => imgui.io_mut().key_map[1] = pressed as u32, //FIXME investigate how this mapping is working now
                    Some(VirtualKeyCode::Z) => imgui.io_mut().key_map[1] = pressed as u32, //FIXME investigate how this mapping is working now
                    Some(VirtualKeyCode::LControl) | Some(VirtualKeyCode::RControl) => imgui.io_mut().key_ctrl = pressed,
                    Some(VirtualKeyCode::LShift) | Some(VirtualKeyCode::RShift) => imgui.io_mut().key_shift = pressed,
                    Some(VirtualKeyCode::LAlt) | Some(VirtualKeyCode::RAlt) => imgui.io_mut().key_alt = pressed,
                    Some(VirtualKeyCode::LWin) | Some(VirtualKeyCode::RWin) => imgui.io_mut().key_super = pressed,
                    _ => {}
                }
            }
            CursorMoved { position: xy, .. } => mouse_state.pos = xy.into(),
            MouseInput { state, button, .. } => {
                match button {
                    MouseButton::Left => mouse_state.pressed.0 = state == Pressed,
                    MouseButton::Right => mouse_state.pressed.1 = state == Pressed,
                    MouseButton::Middle => mouse_state.pressed.2 = state == Pressed,
                    _ => {}
                }
            }
            MouseWheel {
                delta: MouseScrollDelta::LineDelta(_x, y),
                phase: TouchPhase::Moved,
                ..
            } => mouse_state.wheel = y,
            MouseWheel {
                delta: MouseScrollDelta::PixelDelta(glutin::dpi::PhysicalPosition{x:_, y}),
                phase: TouchPhase::Moved,
                ..
            } => mouse_state.wheel = y as f32,
            ReceivedCharacter(c) => imgui.io_mut().add_input_character(c),
            _ => (),
        }
    }

    update_mouse(imgui, mouse_state);
}

fn update_mouse(imgui: &mut Context, mouse_state: &mut MouseState) {
    let scale = imgui.io().display_framebuffer_scale;
    imgui.io_mut().mouse_pos = [
        mouse_state.pos.0 as f32 / scale[0],
        mouse_state.pos.1 as f32 / scale[1],
    ];
    imgui.io_mut().mouse_down = [
        mouse_state.pressed.0,
        mouse_state.pressed.1,
        mouse_state.pressed.2,
        false,
        false
    ];
    imgui.io_mut().mouse_wheel = mouse_state.wheel / scale[1];
    mouse_state.wheel = 0.0;
}

use gfx::format::{BlendFormat, RenderFormat, Formatted};

struct RenderImplState {
    #[warn(TODO)]
    //struct State {
        widgets: BTreeMap<u16, (String, u16)>, //TODO add Vec<messages> to every widget
        resources: BTreeMap<ResID, String>,
        objects: BTreeMap<ObjID, (ObjXY, ResID)>,
        hero_x: f32,
        hero_y: f32,
        hf_x: f32,
        hf_y:f32,
    //}
    #[warn(TODO)]
    //struct RenderState {
        tile_colors: HashMap<String,[u8;4]>,
        palette: [[u8; 4]; 256],
        window: WindowedContext<PossiblyCurrent>,
        device: gfx_device_gl::Device,
        factory: gfx_device_gl::Factory,
        main_color: gfx::handle::RenderTargetView<gfx_device_gl::Resources, Rgba8>,
        main_depth: gfx::handle::DepthStencilView<gfx_device_gl::Resources, DepthStencil>,
        encoder: gfx::Encoder<gfx_device_gl::Resources, gfx_device_gl::CommandBuffer>,
        pso_col: gfx::PipelineState<gfx_device_gl::Resources, pipe_col::Meta>,
        pso_tex: gfx::PipelineState<gfx_device_gl::Resources, pipe_tex::Meta>,
        imgui: Context,
        imgui_renderer: Renderer<gfx::format::Rgba8, gfx_device_gl::Resources>,
        angle: f32, //TODO replace by camera.angle
        zoom: f32, //TODO replace by camera.zoom
        w: f64,
        h: f64,
        delta: delta::Delta,
        mouse_state: MouseState, //TODO move to struct Ui
        grids_tiles: Vec<BakedObjTex>,
        grids_owning: Vec<BakedObjTex>,
        grids_heights: Vec<BakedObjCol>,
        dragging: bool,
        last_mouse_x: f64,
        last_mouse_y: f64,
        shift: Vector2<f32>, //TODO replace by camera.{x,y}
        ctrl_pressed: bool,
        threshold: usize,
        show_tiles: bool,
        show_heights: bool,
        show_owning: bool,
        v1: i32,
        //xui: Widgets,
        imgui_want_capture_mouse: bool,
        imgui_want_capture_keyboard: bool,
    //}
}

impl RenderImplState {
    pub fn event (&mut self, event: Event) {
        use render::Wdg;

        /*TODO app.event(event) */
        match event {
            Event::Tiles(tiles) => {
                for tile in tiles {
                    if let Some(color) = self.tile_colors.get(&tile.name) {
                        self.palette[tile.id as usize] = *color;
                    }
                }
            }
            Event::Grid(Surface{ x, y, tiles, z, ol, .. }) => {
                //println!("grid ({}, {})", grid_x, grid_y);
                //TODO app.rebuild_grid_cache(...)
                //XXX FIXME TODO one BIG mesh with all grids in it ?
                //or individual buffer+pipe.data for every grid ?
                //or use texture-per-grid and don't care at all
                let tiles = ObjTex::plane_from_tiles(1100.0, x, y, tiles.as_ref(), &self.palette)
                    .bake(self.main_color.clone(), &mut self.factory);
                self.grids_tiles.push(tiles);
                let owning = ObjTex::plane_from_owning(1100.0, x, y, ol.as_ref())
                    .bake(self.main_color.clone(), &mut self.factory);
                self.grids_owning.push(owning);
                //let heights = ObjCol::grid_from_heights(100, 11.0, grid_x, grid_y, heights.as_ref(), 1.0)
                //    .bake(main_color.clone(), &mut factory, threshold);
                let heights = ObjCol::grid_from_heights2(100, 11.0, x, y, z.as_ref())
                    .bake(self.main_color.clone(), &mut self.factory, self.threshold);
                self.grids_heights.push(heights);
            }
            Event::Obj(id,xy,resid) => {
                //TODO ??? separate static objects like trees and
                //dynamic objects like rabbits to two
                //different caches
                self.objects.insert(id, (xy,resid));
            }
            Event::ObjRemove(ref id) => {
                self.objects.remove(id);
            }
            Event::Hero(ObjXY(x,y)) => {
                //TODO ??? add to objects
                self.hero_x = x as f32;
                self.hero_y = y as f32;
                //FIXME self.shift += Vector(-hero_x,-hero_y);
                self.shift[0] = -self.hero_x;
                self.shift[1] = -self.hero_y;
            }
            Event::Res(id, name) => {
                self.resources.insert(id, name);
            }
            Event::Wdg(Wdg::New(id,name,parent)) => {
                self.widgets.insert(id,(name.clone(),parent));
                //self.xui.add_widget(id,name,parent).expect("unable to ui.add_widget");
            }
            Event::Wdg(Wdg::Msg(_id,_name,_args)) => {
                //self.xui.message(id,(name,args)).expect("unable to ui.message");
            }
            Event::Wdg(Wdg::Del(id)) => {
                self.widgets.remove(&id);
                //self.xui.del_widget(id).expect("unable to ui.del_widget");
            }
            Event::Hearthfire(ObjXY(x,y)) => {
                self.hf_x = x as f32;
                self.hf_y = y as f32;
            }
        }
    }

    pub fn update (&mut self, render_tx: &Sender<driver::Event>, event: &GEvent<()>, should_stop: &mut bool) {
        if *should_stop { return; }
        ui_handle_event(&mut self.imgui, &mut self.mouse_state, event);
        //TODO if ui.handle_event(event) == NOT_HANDLED {
        //         if app.handle_event(event) == NOT_HANDLED {
        //             match (event) {
        //                 ...
        //             }
        //         }
        //}
        if self.imgui_want_capture_mouse || self.imgui_want_capture_keyboard { return; }
        match *event {
            GEvent::WindowEvent { ref event, .. } => {
                //use glutin::ElementState::Pressed;
                use driver::Event::User;
                use driver::UserInput::*;
                match *event {
                    WindowEvent::KeyboardInput { input, .. } => {
                        let pressed = input.state == Pressed;
                        match input.virtual_keycode {
                            Some(VirtualKeyCode::Escape) => {
                                render_tx.send(User(Quit)).expect("unable to send Render::Quit");
                                *should_stop = true;
                            }
                            Some(VirtualKeyCode::LControl) | Some(VirtualKeyCode::RControl) => {
                                self.ctrl_pressed = pressed; //TODO use some kind of keys_state { ... }
                            }
                            Some(VirtualKeyCode::Up) => {
                                render_tx.send(User(Up)).expect("unable to send Render::Up");
                            }
                            Some(VirtualKeyCode::Down) => {
                                render_tx.send(User(Down)).expect("unable to send Render::Down");
                            }
                            Some(VirtualKeyCode::Left) => {
                                render_tx.send(User(Left)).expect("unable to send Render::Left");
                            }
                            Some(VirtualKeyCode::Right) => {
                                render_tx.send(User(Right)).expect("unable to send Render::Right");
                            }
                            _ => {}
                        }
                    }
                    CloseRequested => {
                        render_tx.send(User(Quit)).expect("unable to send Render::Quit");
                        *should_stop = true;
                    }
                    WindowEvent::Resized(glutin::dpi::PhysicalSize{width, height}) => {
                        //TODO app.resize()
                        //TODO ui.resize()
                        self.w = width as f64;
                        self.h = height as f64;
                        gfx_window_glutin::update_views(&self.window, &mut self.main_color, &mut self.main_depth);
                        self.imgui_renderer.update_render_target(self.main_color.clone());
                        //TODO app.update_render_target(main_color.clone())
                        for t in self.grids_tiles.iter_mut() {
                            t.data.out = self.main_color.clone();
                        }
                        for t in self.grids_owning.iter_mut() {
                            t.data.out = self.main_color.clone();
                        }
                        for t in self.grids_heights.iter_mut() {
                            t.data.out = self.main_color.clone();
                        }
                    }
                    MouseWheel { delta: MouseScrollDelta::LineDelta(_, y), .. } => {
                        if self.ctrl_pressed {
                            if y < 0.0 { self.threshold += 1; } else { if self.threshold > 0 { self.threshold -= 1; } }
                        } else {
                            if y < 0.0 { self.zoom *= 1.1; } else { self.zoom *= 0.9; }
                        }
                    }
                    MouseInput {state, button: MouseButton::Left, ..} => self.dragging = state == Pressed,
                    CursorMoved {position: glutin::dpi::PhysicalPosition{x, y}, ..} => {
                        let x = if x < 0.0 { 0.0 } else if x > self.w as f64 { self.w as f64 } else { x };
                        let y = if y < 0.0 { 0.0 } else if y > self.h as f64 { self.h as f64 } else { y };
                        let delta_x = x - self.last_mouse_x;
                        let delta_y = self.last_mouse_y - y; // y-axis is inverted
                        self.last_mouse_x = x;
                        self.last_mouse_y = y;
                        if self.dragging {
                            let rot = Matrix2::from_angle(Rad(-self.angle));
                            let vec = Vector2::new(delta_x as f32, delta_y as f32);
                            self.shift += rot * vec / self.zoom * 2.0;
                            //println!("shift ({} {})", shift[0], shift[1]);
                        }
                    }
                    _ => (),
                }
            }
            _ => (),
        }
    }
}

pub struct RenderImpl {
    event_loop: EventLoop<()>,
    state: RenderImplState,
}

impl RenderImpl {
    pub fn new () -> RenderImpl {

        let event_loop = EventLoop::new();
        let window_builder = glutin::window::WindowBuilder::new()
            .with_title("gfx 2d test".to_string())
            .with_dimensions(glutin::dpi::LogicalSize::new(800.0, 600.0));
        let context_builder = glutin::ContextBuilder::new();

        let (window, device, mut factory, main_color, main_depth) =
            gfx_window_glutin::init::<Rgba8, DepthStencil>(window_builder, context_builder, &event_loop).expect("gfx_window_glutin::init");


        let size = window.get_inner_size().expect("get_inner_size failed");

        let shaders = {
            let version = device.get_info().shading_language;
            if version.is_embedded {
                if version.major >= 3 {
                    Shaders::GlSlEs300
                } else {
                    Shaders::GlSlEs100
                }
            } else {
                if version.major >= 4 {
                    Shaders::GlSl400
                } else if version.major >= 3 {
                    Shaders::GlSl130
                } else {
                    Shaders::GlSl110
                }
            }
        };

        let mut imgui = Context::create();
        imgui.set_ini_filename(None);
        let config = FontConfig::new().oversample_h(1).pixel_snap_h(true).size_pixels(13.0);
        config./*rasterizer_multiply(1.75).*/add_font(&mut imgui.fonts(), include_bytes!("../DejaVuSansMono.ttf"), &FontGlyphRanges::cyrillic());
        config.merge_mode(true).add_default_font(&mut imgui.fonts());

        RenderImpl {
            event_loop: event_loop,
            state: RenderImplState {
                tile_colors: {
                    use ron::de::from_reader;
                    use std::io::BufReader;
                    use std::fs::File;

                    let f = File::open("tile_colors.ron").expect("unable to open tile_colors.ron");
                    from_reader(BufReader::new(f)).expect("unable to deserialize")
                },
                palette: [[0,0,0,255]; 256],
                window: window,
                device: device,
                encoder: factory.create_command_buffer().into(),
                pso_col: factory.create_pipeline_simple(VERTEX_SHADER_COL.as_bytes(), FRAGMENT_SHADER_COL.as_bytes(), pipe_col::new()).expect("create_pipeline_simple failed"),
                pso_tex: factory.create_pipeline_simple(VERTEX_SHADER_TEX.as_bytes(), FRAGMENT_SHADER_TEX.as_bytes(), pipe_tex::new()).expect("create_pipeline_simple failed"),
                imgui_renderer: Renderer::init(&mut imgui, &mut factory, shaders, main_color.clone()).expect("Failed to initialize renderer"),
                imgui: imgui,
                factory: factory,
                main_color: main_color,
                main_depth: main_depth,
                angle: 0.0, //TODO replace by camera.angle
                zoom: 1.0, //TODO replace by camera.zoom
                w: size.width,
                h: size.height,
                delta: delta::Delta::new(),
                mouse_state: MouseState::default(), //TODO move to struct Ui
                grids_tiles: Vec::new(),
                grids_owning: Vec::new(),
                grids_heights: Vec::new(),
                objects: BTreeMap::new(),
                hero_x: 0.0,
                hero_y: 0.0,
                hf_x: 0.0,
                hf_y: 0.0,
                resources: BTreeMap::new(),
                dragging: false,
                last_mouse_x: 0.0,
                last_mouse_y: 0.0,
                shift: Vector2::zero(), //TODO replace by camera.{x,y}
                ctrl_pressed: false,
                threshold: 3,
                show_tiles: true,
                show_heights: true,
                show_owning: false,
                v1: 0,
                widgets: BTreeMap::new(),
                //xui: Widgets::new(),
                imgui_want_capture_mouse: false,
                imgui_want_capture_keyboard: false,
            }
        }
    }

    pub fn init (&self) {
    }

    pub fn event (&mut self, event: Event) {
        self.state.event(event)
    }

    //TODO FIXME split update() into update() and render() ???
    pub fn update (&mut self, render_tx: &Sender<driver::Event>) -> bool {
        // HANDLE EVENTS
        //TODO app.handle(...)
        {
            let mut should_stop = false;
            let state = &mut self.state;
            self.events_loop.poll_events(|event| {
                state.update(render_tx, &event, &mut should_stop);
            });
            if should_stop {
                return false;
            }
        }

        // UPDATE
        let delta_s = self.state.delta.tick();

        //TODO app.update(delta_s);
        //TODO camera.rotate(angle);
        //self.angle += delta_s * 0.1;

        // RENDER
        //TODO app.render(...)
        //FIXME recalc matrices only if something changed (w,h,angle,zoom)
        //TODO let transform = transform(w,h,camera)
        let translate = Matrix3::new(
            1.0,           0.0,           0.0,
            0.0,           1.0,           0.0,
            self.state.shift[0], self.state.shift[1], 1.0,
            );
        let rotate = Matrix3::from_angle_z(Rad(self.state.angle));
        let scale = Matrix3::from_diagonal(Vector3::new(self.state.zoom / self.state.w as f32, self.state.zoom / self.state.h as f32, 1.0));
        let transform = (scale * rotate * translate).into();

        if self.state.show_tiles {
            for t in self.state.grids_tiles.iter_mut() {
                t.data.transform = transform;
            }
        }
        if self.state.show_owning {
            for t in self.state.grids_owning.iter_mut() {
                t.data.transform = transform;
            }
        }
        if self.state.show_heights {
            for t in self.state.grids_heights.iter_mut() {
                t.data.transform = transform;
                t.data.threshold = self.state.threshold as i32;
            }
        }

        self.state.encoder.clear(&self.state.main_color, BLACK);

        if self.state.show_tiles {
            for t in self.state.grids_tiles.iter() {
                self.state.encoder.draw(&t.slice, &self.state.pso_tex, &t.data);
            }
        }

        if self.state.show_owning {
            for t in self.state.grids_owning.iter() {
                self.state.encoder.draw(&t.slice, &self.state.pso_tex, &t.data);
            }
        }

        if self.state.show_heights {
            for t in self.state.grids_heights.iter() {
                self.state.encoder.draw(&t.slice, &self.state.pso_col, &t.data);
            }
        }

        {
            let mut obj = ObjCol::from_objects(&self.state.objects, self.state.hero_x, self.state.hero_y, self.state.hf_x, self.state.hf_y).bake(self.state.main_color.clone(), &mut self.state.factory, self.state.threshold);
            obj.data.transform = transform;
            obj.data.threshold = 0;
            self.state.encoder.draw(&obj.slice, &self.state.pso_col, &obj.data);
        }

        let (width, height) = self.state.window.get_inner_size().expect("unable to get_inner_size").into();
        let ui = self.state.imgui.frame();
        self.state.imgui_want_capture_mouse = ui.want_capture_mouse();
        self.state.imgui_want_capture_keyboard = ui.want_capture_keyboard();

        let fps = (1.0 / delta_s) as usize;
        //FIXME pass &mut RenderImplState instead
        if !run_ui(&ui, fps, self.state.threshold, &mut self.state.show_tiles, &mut self.state.show_heights, &mut self.state.show_owning, &mut self.state.v1, &self.state.objects, &self.state.resources) {
            return false;
        }
        self.state.imgui_renderer.render(ui, &mut self.state.factory, &mut self.state.encoder).expect("IMGUI Rendering failed");

        self.state.encoder.flush(&mut self.state.device);
        self.state.window.swap_buffers().expect("window.swap_buffers() failed");
        self.state.device.cleanup();

        true
    }

    pub fn end (self) {
    }
}
